from django.conf import settings
from django.test import override_settings
from django_gcp.exceptions import UnknownActionError

from ._base import BaseCommand


class Command(BaseCommand):
    """Use `python manage.py help create_resources` to display help for this command line administration tool"""

    help = "Allows creation of scheduler jobs and pubsub subscriptions via django-gcp"

    def add_arguments(self, parser):

        parser.add_argument(
            "actions",
            nargs="+",
            help="List of task manager actions to take. Valid actions include 'create_scheduler_jobs' and 'create_pubsub_subscriptions'",
        )

        parser.add_argument(
            "--cleanup",
            action="store_true",
            default=False,
            help="Clean up unused resources whose name is affixed with GCP_TASKS_RESOURCE_AFFIX",
        )

        parser.add_argument(
            "--tasks-domain",
            type=str,
            help="Optionally specify a domain to which the tasks will be sent. Overrides the default GCP_TASKS_DOMAIN value.",
        )

    def handle(self, actions, **options):

        cleanup = options["cleanup"]
        tasks_domain = options["tasks_domain"] or settings["GCP_TASKS_DOMAIN"]

        with override_settings(GCP_TASKS_DOMAIN=tasks_domain):

            for action in actions:
                if action == "create_scheduler_jobs":
                    updated, deleted = self.task_manager.create_scheduler_jobs(cleanup=cleanup)
                    report = [f"[+] {name}" for name in updated] + [f"[-] {name}" for name in deleted]
                    self.display_task_report(report, "create", "scheduler jobs")

                elif action == "create_pubsub_subscriptions":
                    updated, deleted = self.task_manager.create_pubsub_subscriptions(cleanup=cleanup)
                    report = [f"[+] {name}" for name in updated] + [f"[-] {name}" for name in deleted]
                    self.display_task_report(report, "create", "pubsub subscriptions")

                else:
                    raise UnknownActionError(
                        f"Unknown action {action}. Use `python manage.py task_manager --help` to see all options"
                    )



import json
from django.conf import settings
from django.forms import Widget


DEFAULT_ACCEPT_MIMETYPE = "*/*"


class CloudObjectWidget(Widget):
    """A widget allowing upload of a file directly to cloud storage

    Must be placed in a form context. On submission of the form,
    upload is started. The submission is intercepted and re-triggered
    on completion of the upload.
    """

    class Media:
        """Add css and js assets to the template rendering the widget"""

        css = {
            "all": ["django_gcp/cloud_object_widget.css"],
        }
        js = ["django_gcp/cloud_object_widget.js"]

    template_name = "django_gcp/cloud_object_widget.html"

    def __init__(
        self,
        *args,
        signed_ingress_url=None,
        ingress_path=None,
        accept_mimetype=DEFAULT_ACCEPT_MIMETYPE,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs
        self.accept_mimetype = accept_mimetype
        self.signed_ingress_url = signed_ingress_url
        self.ingress_path = ingress_path

        if "unfold" in settings.INSTALLED_APPS:
            self.template_name = "django_gcp/contrib/unfold/cloud_object_widget.html"

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        # if self.is_initial():
        #     context["my_context"] = "Adding a new model"
        # else:
        #     context["my_context"] = "Changing an existing model"

        context["signed_ingress_url"] = self.signed_ingress_url
        context["ingress_path"] = self.ingress_path
        context["accept_mimetype"] = self.accept_mimetype
        # Yes I know this is a hack but I've been working on this for weeks and django's a nightmare
        value_or_dict = json.loads(value) or {}
        context["existing_path"] = value_or_dict.get("path", "")
        # context["signing_token"] = a token
        # context["signing_url"] = submit the token to this url in exchange for a signed URL reverse("gcp-storage-get-direct-upload-url")
        return context



# Disabled because the handler must capture in the event of all exceptions
# pylint: disable=broad-except

from logging import StreamHandler
from django.conf import settings
from google.cloud import error_reporting


class GoogleErrorReportingHandler(StreamHandler):
    """A logging handler which streams logs to google cloud error handling.
    Set the GCP_ERROR_REPORTING_SERVICE_NAME environment variable so that GCP Error Reporting allows
    you to filter logs down to those coming from to this service.
    """

    def __init__(self, *args, **kwargs):
        try:
            self.custom_handler_client = error_reporting.Client(service=self.error_reporting_service_name)

        except Exception as exp:
            print(
                "GoogleErrorReportingHandler client not connected, possibly due to invalid or missing google application credentials. Error was:"
            )
            print(exp)
            print("Skipping initialisation of Error Reporting handler.")
            self.custom_handler_client = None
        super().__init__(*args, **kwargs)

    def emit(self, record):
        try:
            # Apply any formatters
            msg = self.format(record)

            # Send report to GCP Error Reporting
            self.custom_handler_client.report(msg)

        # Avoid exception logging on keyboard error or system exit
        # pylint: disable-next=try-except-raise
        except (KeyboardInterrupt, SystemExit):
            raise

        except Exception as e:
            self.handleError(record)
            if self.custom_handler_client:
                self.custom_handler_client.report_exception()
            self.stream.write("Exception in custom logging handler - GoogleErrorReportingHandler")
            self.stream.write(e)
            self.flush()

    @property
    def error_reporting_service_name(self):
        """The name of this service, as visible in GCP error reporting

        This is determined from the GCP_ERROR_REPORTING_SERVICE_NAME setting. If
        you're running in GCP Cloud Run, then it's recommended to use `env.HOSTNAME`
        to set this value.

        """
        return getattr(settings, "GCP_ERROR_REPORTING_SERVICE_NAME")



import datetime
import logging
from django.utils import timezone
from django_gcp.exceptions import AttemptedOverwriteError, MissingBlobError
from google.cloud.exceptions import NotFound, PreconditionFailed


logger = logging.getLogger(__name__)


def blob_exists(bucket, blob_name):
    """Quick check that a blob with a given name exists in a bucket"""
    blob = bucket.blob(blob_name)
    return blob.exists()


def copy_blob(
    source_bucket,
    source_blob_name,
    destination_bucket,
    destination_blob_name,
    overwrite=False,
    move=False,
    attributes=None,
):
    """Copies or moves a blob from one bucket to another with a new name.

    :param source_bucket: The google.cloud.storage.bucket object you're copying from
    :param source_blob_name: The name of the GCS object to be copied eg "path/in/bucket/source-blob-name.txt"
    :param destination_bucket_name: The google.cloud.storage.bucket object you're copying to
    :param destination_blob_name: The name of the new GCS object eg "path/in/bucket/destination-blob-name.txt"
    :param overwrite: If True, allows an existing destination file to be overwritten
    :param move: If True, removes the source blob after copy (ie moves it rather than duplicating it)
    :param attributes: A dict of values used to update attributes of the destination blob (eg {"content_type": "image/png"})
    """

    source_blob = source_bucket.blob(source_blob_name)

    extra_args = {}
    if not overwrite:
        # Optional: set a generation-match precondition to avoid potential race conditions
        # and data corruptions. The request is aborted if the object's
        # generation number does not match your precondition. For a destination
        # object that does not yet exist, set the if_generation_match precondition to 0.
        # If the destination object already exists in your bucket, set instead a
        # generation-match precondition using its generation number.
        # There is also an `if_source_generation_match` parameter, which is not used in this example.
        destination_generation_match_precondition = 0
        extra_args["if_generation_match"] = destination_generation_match_precondition

    try:
        destination_blob = source_bucket.copy_blob(source_blob, destination_bucket, destination_blob_name, **extra_args)

        verb = "copied"
        if move:
            verb = "moved"
            source_bucket.delete_blob(source_blob_name)

        logger.info(
            "Blob %s in bucket %s %s to blob %s in bucket %s",
            source_blob_name,
            source_bucket.name,
            verb,
            destination_blob_name,
            destination_bucket.name,
        )

        if attributes is not None:
            for key, value in attributes.items():
                setattr(destination_blob, key, value)
                logger.debug("Set attribute %s to %s on destination blob %s", key, value, destination_blob_name)

            destination_blob.patch()
            logger.info("Attributes of blob %s updated", destination_blob_name)

        return destination_blob

    except NotFound as e:
        raise MissingBlobError(
            f"Could not complete copy: source blob {source_blob_name} does not exist in bucket {source_bucket.name}"
        ) from e

    except PreconditionFailed as e:
        raise AttemptedOverwriteError(
            f"Could not complete copy: blob {destination_blob_name} in bucket {destination_bucket.name} already exists (source {source_blob_name} in bucket {source_bucket.name})"
        ) from e


def delete_blob(bucket, blob_name, generation=None, ignore_missing=False):
    """Deletes a blob, with the ability to handle missing blobs

    :param bucket: The bucket object from which the blob will be deleted
    :param blob_name: The name of the blob to be deleted eg "path/in/bucket/your-object-name.txt"
    :param generation: the specific generation of a versioned blob to delete. Unless generation is passed, deletion of versioned blobs will
    simply add a time_deleted parameter to those blobs.
    :param ignore_missing: If True, attempting to delete a missing blob will pass without raising an exception.
    This option should only be used when there's a real chance that objects may already have been cleaned up
    by another process by the time you call this (to prevent race conditions). Otherwise leave False since
    failure at this point can be indicative of upstream errors in handling that blob.
    """
    try:
        bucket.delete_blob(blob_name, generation=generation)
        logger.info("Deleted blob %s from bucket %s", blob_name, bucket.name)

    except NotFound as e:
        if not ignore_missing:
            raise MissingBlobError("Could not delete blob %s from bucket %s - blob not found") from e

        logger.info("Attempted to delete blob %s from bucket %s - blob missing", blob_name, bucket.name)


def get_generations(bucket, blob_name):
    """Get blobs corresponding to all generations of an object
    TODO Work up a more useful output than simply a list of objects
    :param bucket: The bucket object from which the blob will be deleted
    :param blob_name: The name of the blob to be deleted eg "path/in/bucket/your-object-name.txt"
    """

    return list(bucket.client.list_blobs(bucket.name, versions=True, prefix=blob_name))


def get_signed_upload_url(bucket, blob_name, timedelta=None, max_size_bytes=None, **kwargs):
    """Get a signed URL for uploading a blob to GCS

    :param google.cloud.storage.Bucket bucket: The bucket to which the blob will be uploaded
    :param str blob_name: Name of the blob within the bucket
    :param Union[datetime.timedelta, None] timedelta: A datetime.timedelta object representing the
    period which the upload URL should be valid for (default is 60 minutes)
    :param Union[int, None] max_size_bytes: Maximum size allowable, added to any headers that are supplied in kwargs.
    """
    if timedelta is None:
        timedelta = datetime.timedelta(minutes=60)

    blob = bucket.blob(blob_name)
    if max_size_bytes is not None:
        content_length_range = f"0,{max_size_bytes}"
        headers = kwargs.pop("headers", {})
        headers["X-Goog-Content-Length-Range"]: content_length_range

    return blob.generate_signed_url(expiration=timezone.now() + timedelta, method="PUT", **kwargs)



import asyncio
from asgiref.sync import async_to_sync


def run_coroutine(handler, **kwargs):
    """Run an async routine, by converting to synchronous if possible"""
    try:
        return async_to_sync(handler)(**kwargs)
    except RuntimeError:
        coroutine = handler(**kwargs)
        return asyncio.get_event_loop().create_task(coroutine)



# Signal handlers have to be imported on app ready state to avoid circular imports
# pylint: disable=import-outside-toplevel

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoGCPAppConfig(AppConfig):
    """Django application metadata and config"""

    name = "django_gcp"
    label = "django_gcp"
    verbose_name = _("Django GCP")
    task_manager = None

    def ready(self):
        from django_gcp.signals import register_signal_handlers

        register_signal_handlers()

        from django_gcp.tasks import TaskManager

        self.task_manager = TaskManager()
